class Record(dict):

  def __init__(self, **dic):
    self.update(dic)

  def __getattr__(self, i):
    if i in self:
      return self[i]
    else:
      return super(self.__class__, self).__getattr__(i)

  def __setattr__(self,i , v):
    self[i] = v
    return v

  def __repr__(self):
    attrs = []
    if 'id' in self:
      attrs += ['id=%s' % self.id]
    for att_name, attr_value in self.iteritems():
      if att_name == 'id': continue
      attrs += ['%s=%s' % (att_name, attr_value.__repr__())]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(attrs))



import psycopg2 as pg
from psycopg2 import extras
import urlparse
from query_builders import SelectQueryBuilder, InsertQueryBuilder, UpdateQueryBuilder, DeleteQueryBuilder, RawQueryBuilder
import config
from tools import retry


class DatabaseConnections(object):

  _connections = {}
  _urls = {}

  @classmethod
  def connection(self, db_name):
    if db_name not in self._connections:
      self._connections[db_name] = self.build_connection(db_name)
    return self._connections[db_name]

  @classmethod
  def build_connection(self, name):
    db = self.urls(name)
    return DatabaseConnection(db)

  @classmethod
  def urls(self, name):
    if len(self._urls) == 0:
      config.init()
      for nme, url in config.DATABASES.iteritems():
        if url:
          self._urls[nme] = urlparse.urlparse(url)
    return self._urls[name]


class DatabaseConnection(object):

  _connection = None
  _cursor = None

  def __init__(self, db_config):
    self.db_config = db_config

  @retry(pg.OperationalError, tries=3)
  def connect(self):
    self._cursor = None
    connection = pg.connect(
      connection_factory=extras.MinTimeLoggingConnection,
      database=self.db_config.path[1:],
      user=self.db_config.username,
      password=self.db_config.password,
      host=self.db_config.hostname,
      port=self.db_config.port,
      connect_timeout=5)
    connection.initialize(config.logger)
    connection.autocommit = True
    return connection

  def connection(self):
    if self._connection is None:
      self._connection = self.connect()
    return self._connection

  def cursor(self):
    if self._cursor is None:
      self._cursor = self.connection().cursor(cursor_factory = pg.extras.RealDictCursor)
    return self._cursor

  @retry((pg.InterfaceError, pg.extensions.TransactionRollbackError), tries=3)
  def execute(self, *query):
    try:
      return self.cursor().execute(*query)
    except pg.InterfaceError:
      self._connection = None
      self._cursor = None
      raise


class DatabaseAdapter(object):

  def __init__(self, db, model_metadata):
    self.db = db
    self.model_metadata = model_metadata

  def select(self, **kwargs):
    kwargs['model_metadata'] = self.model_metadata
    query = SelectQueryBuilder(**kwargs).query
    config.logger.debug(query)
    self.db.execute(*query)
    return self.db.cursor().fetchall(), self.columns()

  def insert(self, **kwargs):
    kwargs['model_metadata'] = self.model_metadata
    query = InsertQueryBuilder(**kwargs).query
    config.logger.debug(query)
    self.db.execute(*query)
    row_id = self.db.cursor().fetchone()['id']
    return self.select(where = {'id': row_id})

  def update(self, **kwargs):
    kwargs['model_metadata'] = self.model_metadata
    query = UpdateQueryBuilder(**kwargs).query
    config.logger.debug(query)
    self.db.execute(*query)
    row_id = self.db.cursor().fetchone()['id']
    return self.select(where = {'id': row_id})

  def delete(self, **kwargs):
    kwargs['model_metadata'] = self.model_metadata
    query = DeleteQueryBuilder(**kwargs).query
    config.logger.debug(query)
    self.db.execute(*query)

  def raw_query(self, **kwargs):
    query = RawQueryBuilder(**kwargs).query
    config.logger.debug(query)
    self.db.execute(*query)
    return self.db.cursor().fetchall(), self.columns()

  def columns(self):
    return [col_desc[0] for col_desc in self.db.cursor().description]



from memoized_property import memoized_property
from datetime import datetime
import pandas as pd
import numpy as np
import model, config, re, collections, json


class PGQueryBuilder(object):

  def __init__(self, **kwargs):
    self.kwargs = kwargs

  @memoized_property
  def model_metadata(self):
    return self.kwargs['model_metadata']

  @memoized_property
  def table_name(self):
    return self.model_metadata['table_name']

  @memoized_property
  def table_alias(self):
    return self.model_metadata['table_alias']

  @memoized_property
  def belongs_to(self):
    return self.model_metadata['belongs_to']

  @memoized_property
  def scopes(self):
    return self.model_metadata['scopes']

  @memoized_property
  def now(self):
    return datetime.utcnow()

  def extrapolators(self, fields, sep = ', '):
    extrapolators = []
    for field in fields: extrapolators.append('%(' + '%s' % field + ')s')
    return sep.join(extrapolators)

  @memoized_property
  def stack(self):
    return self.kwargs.get('stack', '')

  @memoized_property
  def watermark(self):
    return "/*%s | %s */ " % (config.WATERMARK, self.stack)


class SelectQueryBuilder(PGQueryBuilder):

  where_values_prefix = ''
  where_values = {}

  @memoized_property
  def selects(self):
    selects = self.kwargs.get('select', '*')
    if isinstance(selects, str):
      return selects
    elif isinstance(selects, list):
      return ', '.join(selects)

  @memoized_property
  def froms(self):
    if self.table_alias is not None:
      return "%(table_name)s %(table_alias)s" % {'table_name': self.table_name, 'table_alias': self.table_alias}
    else:
      return self.table_name

  @memoized_property
  def scope_wheres(self):
    scopes = self.kwargs.get('scopes', [])
    if not isinstance(scopes, list): scopes = [scopes]
    results = []
    for scope in scopes:
      scp = self.scopes[scope]
      if not isinstance(scp, list): scp = [scp]
      results += scp
    return results

  @memoized_property
  def wheres(self):
    self.where_values = {}
    wheres = self.kwargs.get('where', None)
    if not isinstance(wheres, list): wheres = [wheres]
    wheres += self.scope_wheres
    res = [self.where_items(where) for where in wheres]
    results = [item for sublist in res for item in sublist]
    return ' AND '.join(results)

  def add_to_where_values(self, values):
    for k, v in values.iteritems():
      if isinstance(v, pd.Series) or isinstance(v, list):
        v = tuple(v)
      self.where_values[self.where_label(k)] = v

  def where_label(self, label):
    return self.where_values_prefix + label

  def where_items(self, where):
    results = []
    if isinstance(where, str):
      results += [where]
    elif isinstance(where, tuple):
      self.add_to_where_values(where[1])
      results += [where[0]]
    elif isinstance(where, dict):
      for k, v in where.iteritems():
        if isinstance(v, tuple):
          from_label = self.where_label(k + '_from')
          to_label = self.where_label(k + '_to')
          results += [k + ' BETWEEN %(' + from_label + ')s AND %(' + to_label + ')s']
          self.add_to_where_values({from_label: v[0], to_label: v[1]})
        elif isinstance(v, dict):
          for kk, vv in v.iteritems():
            res = "(" + k + "->>'" + kk + "')"
            if isinstance(vv, int):
              res += '::INTEGER'
            elif isinstance(vv, float):
              res += '::FLOAT'
            label = k + '_' + kk
            res += " = %(" + self.where_label(label) + ")s"
            results += [res]
            self.add_to_where_values({label: vv})
        elif not isinstance(v, list) and not isinstance(v, pd.Series) and not isinstance(v, np.ndarray) and pd.isnull(v):
          results += [k + ' IS NULL']
        else:
          self.add_to_where_values({k: v})
          if isinstance(v, list) or isinstance(v, pd.Series):
            operator = 'IN'
          else:
            operator = '='
          results += [k + ' ' + operator + ' %(' + self.where_label(k) + ')s']
    elif isinstance(where, list):
      self.add_to_where_values(where[1])
      result = where[0]
      for l in re.findall('%\((\S+)\)s', result):
        result = re.sub('%\(' + l + '\)s', '%(' + self.where_label(l) + ')s', result)
      results += [result]
    return results

  @memoized_property
  def limit(self):
    return self.kwargs.get('limit', None)

  @memoized_property
  def order_bys(self):
    return self.kwargs.get('order', None)

  @memoized_property
  def group_bys(self):
    return self.kwargs.get('group', None)

  @memoized_property
  def left_joins(self):
    joins = self.kwargs.get('left_join', None)
    return self.joins(joins, 'LEFT')

  @memoized_property
  def inner_joins(self):
    joins = self.kwargs.get('inner_join', None)
    if joins is None: return
    if isinstance(joins, str):
      return "INNER JOIN %s" % joins
    elif isinstance(joins, list):
      js = []
      for j in joins:
        if isinstance(j, str):
          js += ["INNER JOIN %s" % j]
        elif issubclass(j, model.Model):
          js += [self.build_join(j, how = 'INNER')]
      return ' '.join(js)

  def joins(self, joins, how):
    if joins is None: return
    if isinstance(joins, str):
      return "%s JOIN %s" % (how, joins)
    elif isinstance(joins, list):
      js = []
      for j in joins:
        if isinstance(j, str):
          js += ["%s JOIN %s" % (how, j)]
        elif issubclass(j, model.Model):
          js += [self.build_join(j, how = how)]
      return ' '.join(js)

  def build_join(self, join_model, how = 'INNER'):
    join_model = join_model.model_metadata()
    table_name, join_table_name = self.table_name, join_model['table_name']
    table_alias, join_table_alias = self.table_alias, join_model['table_alias']
    if self.table_name in join_model['belongs_to']:
      foreign_key = join_model['belongs_to'][table_name]
      primary_key = 'id'
    else:
      primary_key = self.model_metadata['belongs_to'][join_table_name]
      foreign_key = 'id'
    return "%(how)s JOIN %(join_table_name)s %(join_table_alias)s ON %(table_alias)s.%(primary_key)s = %(join_table_alias)s.%(foreign_key)s" % {'how': how, 'join_table_name': join_table_name, 'join_table_alias': join_table_alias, 'table_alias': table_alias, 'foreign_key': foreign_key, 'primary_key': primary_key}

  @memoized_property
  def query(self):
    query = self.watermark + "SELECT " + self.selects + ' FROM ' + self.froms
    if self.left_joins: query += ' ' + self.left_joins
    if self.inner_joins: query += ' ' + self.inner_joins
    if self.wheres: query += ' WHERE ' + self.wheres
    if self.group_bys: query += ' GROUP BY ' + self.group_bys
    if self.order_bys: query += ' ORDER BY ' + self.order_bys
    if self.limit: query += ' LIMIT ' + str(self.limit)
    query += ';'
    return (query, self.where_values)


class WriteQueryBuilder(PGQueryBuilder):

  @memoized_property
  def values(self):
    values = collections.OrderedDict()
    for k, v in self.kwargs['values'].iteritems():
      if isinstance(v, dict):
        v = json.dumps(v)
      values[k] = v
    values['updated_at'] = self.now
    if 'stack' in values: del values['stack']
    return values

  @memoized_property
  def field_array(self): return self.values.keys()

  @memoized_property
  def fields(self):
    return ', '.join(self.field_array)


class InsertQueryBuilder(WriteQueryBuilder):

  @memoized_property
  def values(self):
    values = super(InsertQueryBuilder, self).values
    values['created_at'] = self.now
    return values

  @memoized_property
  def query(self):
    query = self.watermark + "INSERT INTO " + self.table_name + " (" + self.fields + ") VALUES (" + self.extrapolators(self.field_array, sep = ', ') + ") RETURNING id;"
    return (query, self.values)


class UpdateQueryBuilder(WriteQueryBuilder, SelectQueryBuilder):

  where_values_prefix = 'w_'

  @memoized_property
  def query(self):
    query = self.watermark + 'UPDATE ' + self.table_name + ' SET (' + self.fields + ') = (' + self.extrapolators(self.field_array, sep = ', ') + ')'
    if self.wheres: query += " WHERE " + self.wheres
    query += ' RETURNING id;'
    values = self.where_values
    values.update(self.values)
    return (query, values)


class DeleteQueryBuilder(WriteQueryBuilder, SelectQueryBuilder):

  @memoized_property
  def query(self):
    query = self.watermark + 'DELETE FROM ' +self.table_name + ' WHERE ' + self.wheres + ';'
    return (query, self.where_values)


class RawQueryBuilder(WriteQueryBuilder, SelectQueryBuilder):

  @memoized_property
  def sql(self):
    if 'sql' in self.kwargs and self.kwargs['sql']:
      raw_sql = self.kwargs['sql']
    if 'filename' in self.kwargs and self.kwargs['filename']:
      with open(self.kwargs['filename']) as file:
        raw_sql = file.read()
    return re.sub(r'\{(\w+?)\}', r'%(\1)s', raw_sql)

  @memoized_property
  def query(self):
    query = self.watermark + self.sql
    self.wheres
    return (query, self.where_values)



from database import DatabaseAdapter, DatabaseConnections
from record import Record
import pandas as pd
import numpy as np
import re, inspect
import os

class ModelIterator(object):

  def __init__(self, model):
    self.model = model
    self.current = 0

  def __iter__(self): return self

  def next(self):
    if self.current < len(self.model):
      record = self.model.record_class(**self.model.iloc[self.current])
      self.current += 1
      return record
    else:
      raise StopIteration()

class Model(pd.DataFrame):
  ROLES = {"replica": "read", "master": "write"}  # to have backward compatibility
  table_name = None
  table_alias = None
  db_names = {} # {'master': database_master_url, 'replica': database_replica_url}
  has_many = []
  belongs_to = {}
  scopes = {}
  record_class = Record

  def __init__(self, *args, **kwargs):
    self.create_relationships()
    self.count = self._instance_count
    self.insert = self._instance_insert
    self.update = self._instance_update
    super(Model, self).__init__(*args, **kwargs)

  @property
  def _constructor(self):
    return self.__class__

  @classmethod
  def create_relationships(self):
    this_table_name = self.model_metadata()['table_name']
    for h in self.has_many:
      other_table_name = h.model_metadata()['table_name']
      def func(self):
        return h.select(where={h.belongs_to[this_table_name]: self.id})
      setattr(self, other_table_name, func)

  @classmethod
  def instance(self, result):
    return self.from_records(result[0], columns=result[1], coerce_float=True)

  @classmethod
  def stack_mark(self, stack):
    filename = stack[1][1]
    function_name = stack[1][3]
    line_number = stack[1][2]
    return ':'.join([filename, function_name, str(line_number)])

  @classmethod
  def select(self, **kwargs):
    kwargs['stack'] = self.stack_mark(inspect.stack())
    return self.instance(self.db_adapter(db_name=kwargs.get('db'), role=kwargs.get('role', 'replica')).select(**kwargs))

  @classmethod
  def query(self, sql=None, filename=None, **kwargs):
    """ run raw sql from sql or file """
    kwargs['stack'] = self.stack_mark(inspect.stack())
    if filename:
      filename = os.path.join(os.environ['PWD'], filename)
    kwargs['where'] = kwargs.get('where', kwargs.get('params'))
    return self.instance(self.db_adapter(db_name=kwargs.get('db'), role=kwargs.get('role', 'replica')).raw_query(sql=sql, filename=filename, **kwargs))

  @classmethod
  def count(self, **kwargs):
    kwargs['select'] = 'COUNT(*)'
    return self.db_adapter(db_name=kwargs.get('db'), role=kwargs.get('role', 'replica')).select(**kwargs)[0][0]['count']

  def _instance_count(self, **kwargs): return super(Model, self).count(**kwargs)

  @classmethod
  def insert(self, **kwargs):
    kwargs['stack'] = self.stack_mark(inspect.stack())
    return self.record_class(**self.db_adapter(role='master').insert(**kwargs)[0][0])

  def _instance_insert(self, *args, **kwargs): return super(Model, self).insert(*args, **kwargs)

  @classmethod
  def update(self, **kwargs):
    kwargs['stack'] = self.stack_mark(inspect.stack())
    return self.instance(self.db_adapter(role='master').update(**kwargs))

  def _instance_update(self, *args, **kwargs): return super(Model, self).update(*args, **kwargs)

  @classmethod
  def delete(self, **kwargs):
    kwargs['stack'] = self.stack_mark(inspect.stack())
    return self.db_adapter(role='master').delete(**kwargs)

  @classmethod
  def last(self, limit=1, **kwargs):
    return self.instance(self.db_adapter(db_name=kwargs.get('db'), role=kwargs.get('role', 'replica')).select(where='created_at IS NOT NULL', order='created_at DESC', limit=limit))

  @classmethod
  def find_by(self, values={}, **kwargs):
    return self.record_class(**self.db_adapter(db_name=kwargs.get('db'), role=kwargs.get('role', 'replica')).select(where=values, limit=1)[0][0])

  @classmethod
  def find(self, id, **kwargs):
    return self.find_by(values={'id': id}, **kwargs)

  @classmethod
  def db_adapter(self, role='replica', db_name=None):
    if not hasattr(self, '_db_adapter'): self._db_adapter = {}
    if role not in self._db_adapter:
      self._db_adapter[role] = DatabaseAdapter(self.db(role=role, db_name=db_name), self.model_metadata())
    return self._db_adapter[role]

  @classmethod
  def model_metadata(self):
    tn = self.table_name if isinstance(self.table_name, str) else self.default_table_name()
    table_alias = self.table_alias
    if table_alias is None: table_alias = ''.join([w[0] for w in tn.split('_')])
    return {
      'table_name': tn,
      'table_alias': table_alias,
      'belongs_to': self.belongs_to,
      'scopes': self.scopes
    }

  @classmethod
  def default_table_name(self):
    if '_default_table_name' not in self.__dict__:
      name = self.__name__
      s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
      self._default_table_name = re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()
    return self._default_table_name

  @classmethod
  def db(self, role='replica', db_name = None):
    if not hasattr(self, '_db'): self._db = {}
    name = self.db_names.get(role, self.db_names.get(self.ROLES.get(role))) if db_name is None else db_name
    if name not in self._db: self._db[name] = DatabaseConnections.connection(name)
    return self._db[name]

  @classmethod
  def _use_replica(self, **kwargs):
    try:
      kwargs['stack'] = self.stack_mark(inspect.stack())
      sql = "select setting FROM pg_settings WHERE name = 'hot_standby'"
      r = self.instance(self.db_adapter().raw_query(sql=sql, **kwargs)).squeeze()
      return r == "on"
    except:
      return False

  @classmethod
  def replica_lag(self, **kwargs):
    if not self._use_replica():
      return 0
    try:
      kwargs['stack'] = self.stack_mark(inspect.stack())
      sql = "select EXTRACT(EPOCH FROM NOW() - pg_last_xact_replay_timestamp()) AS replication_lag"
      return self.instance(self.db_adapter().raw_query(sql=sql, **kwargs)).squeeze()
    except:
      return 0

  def where(self, **kwargs):
    conditions = kwargs.get('conditions', kwargs)
    if 'where_not' in conditions:
      wnot = conditions['where_not']
      del conditions['where_not']
    else:
      wnot = False
    filt = True
    for field, value in conditions.iteritems():
      if value is None:
        nf = pd.isnull(self[field])
      elif isinstance(value, (np.ndarray, list, pd.Series)):
        nf = self[field].isin(value)
      else:
        nf = self[field] == value
      if wnot: nf = ~nf
      filt = filt & nf
    return self[filt]

  def where_not(self, **kwargs):
    kwargs['where_not'] = True
    return self.where(**kwargs)

  def records(self):
    return ModelIterator(self)



from model import Model
from record import Record
import config

__author__ = 'Emmanuel Turlay'
__copyright__ = 'Copyright 2017, Instacart'
__credits__ = ['Emmanuel Turlay', 'Mathieu Ripert']
__license__ = 'MIT'
__version__ = '0.5.2'
__maintainer__ = 'Emmanuel Turlay'
__email__ = 'emmanuel@instacart.com'
__status__ = 'Prototype'


from setuptools import setup
from os import path

import jardin

here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.rst')) as f:
    long_description = f.read()

setup(name = 'jardin',
      version = jardin.__version__,
      description = 'A dataframe-base ORM',
      long_description = long_description,
      url = 'https://github.com/instacart/jardin',
      author = 'Emmanuel Turlay',
      license = 'MIT',
      author_email = 'emmanuel@instacart.com',
      packages = ['jardin'],
      install_requires = [
      'pandas',
      'numpy',
      'psycopg2'
      ],
      python_requires='>=2.7, <3',
      classifiers = [
      'Intended Audience :: Developers',
      'License :: OSI Approved :: MIT License',
      'Programming Language :: Python :: 2.7',
      ],
      keywords = 'postgres database ORM')